/*****************************************************************************
 *
 * wolcub, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ebook@ndl.kiev.ua>
 *
 * Heavily based on BITMAP OEM DLL example from WDK by Microsoft DDK Team.
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

//  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
//  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//  PARTICULAR PURPOSE.
//
//  Copyright  1998 - 2003  Microsoft Corporation.  All Rights Reserved.
//
//  FILE:    DDIHook.cpp
//    
//
//  PURPOSE:  Implementation of DDI Hook OEMEndDoc. This function
//          dumps the buffered bitmap data out. 
//
//
//  Functions:
//          OEMEndDoc
//
//      
//
//
//  PLATFORMS:  Windows XP, Windows Server 2003, Windows codenamed Longhorn
//
//
//  History: 
//          06/24/03    xxx created.
//
//

#include "precomp.h"
#include "ddihook.h"
#include "wolcub.h"
#include "debug.h"
#include "wolutil.h"
#include "magick/MagickCore.h"

void ProcessImage(lUInt8* bitmap, DWORD& dwBufSize, DWORD& dwNewImgHeight,
                  DWORD width, DWORD height, BYTE btBPP,
                  BOOL bLandscape, BOOL bSplit, WOLWriter* pWolWriter,
                  BOOL bFlush);

BOOL APIENTRY
  OEMStartDoc(
    SURFOBJ  *pso,
    PWSTR  pwszDocName,
    DWORD  dwJobId
    )
{
    OEMDBG(DBG_VERBOSE, L"OEMStartDoc entry.");

    PDEVOBJ pDevObj = (PDEVOBJ)pso->dhpdev;
    POEMPDEV pOemPDEV = (POEMPDEV)pDevObj->pdevOEM;

    // Get the temp path.
    TCHAR tmpPath[_MAX_PATH];
    DWORD dwRetVal = GetTempPath(_MAX_PATH, tmpPath);
    if (dwRetVal > _MAX_PATH || (dwRetVal == 0))
    {
        return FALSE;
    }

    // Create a temporary file. 
    UINT uRetVal = GetTempFileName(tmpPath,                 // directory for tmp files
                                   _T("wol"),               // temp file name prefix
                                   0,                       // create unique name
                                   pOemPDEV->pTmpFilePath); // buffer for name
    if (uRetVal == 0)
    {
        return FALSE;
    }

    pOemPDEV->lvRefTmpStream = LVOpenFileStream(pOemPDEV->pTmpFilePath, LVOM_READWRITE);
    if (!pOemPDEV->lvRefTmpStream.get())
    {
        return FALSE;
    }

    pOemPDEV->pWolWriter = new WOLWriter(pOemPDEV->lvRefTmpStream.get());

    pOemPDEV->pWolWriter->addTitle(lString8("Empty Title"),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""),
                                   lString8(""));

    MagickCoreGenesis(NULL, MagickFalse);

    return (pOemPDEV->m_pfnDrvStartDoc)(pso, pwszDocName, dwJobId);
}

BOOL APIENTRY
  FinishPrevPage(
    SURFOBJ  *pso, BOOL bFlush
    )
{
    OEMDBG(DBG_VERBOSE, L"FinishPage entry.");

    PDEVOBJ pDevObj = (PDEVOBJ)pso->dhpdev;
    POEMPDEV pOemPDEV = (POEMPDEV)pDevObj->pdevOEM;

    if (pOemPDEV->pBufStart)
    {
        ProcessImage(pOemPDEV->pBufStart, pOemPDEV->dwPrevBufSize,
                     pOemPDEV->dwImgHeight, pOemPDEV->dwImgWidth,
                     pOemPDEV->dwImgHeight, pOemPDEV->btBPP,
                     pOemPDEV->bLandscape, pOemPDEV->bSplit,
                     pOemPDEV->pWolWriter, bFlush);

        pOemPDEV->dwCurBufSize = pOemPDEV->dwPrevBufSize;

        return TRUE;
    }

    return FALSE;
}

BOOL APIENTRY
  OEMStartPage(
    SURFOBJ  *pso
    )
{
    OEMDBG(DBG_VERBOSE, L"OEMStartPage entry.");

    PDEVOBJ pDevObj = (PDEVOBJ)pso->dhpdev;
    POEMPDEV pOemPDEV = (POEMPDEV)pDevObj->pdevOEM;

    FinishPrevPage(pso, FALSE);

    return (pOemPDEV->m_pfnDrvStartPage)(pso); 
}

BOOL APIENTRY
OEMEndDoc(
    SURFOBJ     *pso,
    FLONG       fl
    )

/*++

Routine Description:

    Implementation of DDI hook for DrvEndDoc.

    DrvEndDoc is called by GDI when it has finished 
    sending a document to the driver for rendering.
    
    Please refer to DDK documentation for more details.

Arguments:

    pso - Defines the surface object
    flags - A set of flag bits

Return Value:

    TRUE if successful, FALSE if there is an error

--*/

{
    OEMDBG(DBG_VERBOSE, L"OEMEndDoc entry.");

    PDEVOBJ pDevObj = (PDEVOBJ)pso->dhpdev;
    POEMPDEV pOemPDEV = (POEMPDEV)pDevObj->pdevOEM;
    
    FinishPrevPage(pso, TRUE);

    if (pOemPDEV->lvRefTmpStream.get())
    {
        // Remove WOL Writer first to force update.
        delete pOemPDEV->pWolWriter;
        pOemPDEV->pWolWriter = NULL;

        // Copy temporary file into spool.
        //
        const lvsize_t bufSize = 65536;
        BYTE buf[bufSize];
        pOemPDEV->lvRefTmpStream->Seek(0, LVSEEK_SET, NULL);
        lvsize_t nBytesRead = 0;
        do
        {
            pOemPDEV->lvRefTmpStream->Read(buf, bufSize, &nBytesRead);

            if (nBytesRead)
            {
                pDevObj->pDrvProcs->DrvWriteSpoolBuf(pDevObj, (void*)buf, nBytesRead);
            }
        }
        while(nBytesRead > 0);

        // Free stream and delete temporary file.
        //
        pOemPDEV->lvRefTmpStream.Clear();
        DeleteFile(pOemPDEV->pTmpFilePath);

        vFreeBuffer(pOemPDEV);
    }

    MagickCoreTerminus();

    // Punt call back to UNIDRV.
    //
    return (pOemPDEV->m_pfnDrvEndDoc)(pso, fl);
}
