/*****************************************************************************
 *
 * wolcub, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <ebook@ndl.kiev.ua>
 *
 * Heavily based on BITMAP OEM DLL example from WDK by Microsoft DDK Team.
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "precomp.h"
#include "magick/MagickCore.h"
#include "wolutil.h"
#include "debug.h"

const DWORD IMG_WIDTH = 600, IMG_HEIGHT = 800;
double OVERLAP_RATIO = 0.05;

void FlushImage(Image* image, WOLWriter* pWolWriter)
{
    ExceptionInfo ex;
    DWORD width = image->columns, height = image->rows;

    lUInt8* pixels = new lUInt8[width * height];
    ExportImagePixels(image, 0, 0, width, height, "I", CharPixel, pixels, &ex);

    // Currently lBook V3 supports only 4-level grayscale
    // or monochrome, but we work in 4-level grayscale only.
    DWORD dwPixelsCount = width * height;

    lUInt8* buf = new lUInt8[dwPixelsCount / 4];
    memset(buf, 0, dwPixelsCount / 4);

    for (DWORD i = 0; i < dwPixelsCount; ++i)
        buf[i / 4] |= (((unsigned short)pixels[i] + 42) / 85) << (6 - 2 * (i % 4));

    pWolWriter->addImage(width, height, buf, 2);

    delete[] buf;
    delete[] pixels;
}

Image* GetCroppedImage(Image* image, DWORD& start_y)
{
    OEMDBG(DBG_VERBOSE, L"GetCroppedImage entry.");

    WARNING(DLLTEXT("Entered GetCroppedImage.\n"));

    ExceptionInfo ex;
    Image* res_img = NULL;

    RectangleInfo bb = GetImageBoundingBox(image, &ex);

    if (bb.width > 0 && bb.height > 0)
    {
        WARNING(DLLTEXT("BB is not zero\n"));
        TCHAR fuck[255];
        wsprintf(fuck, L"x: %d, y: %d, width: %d, height: %d, rows: %d\n", bb.x, bb.y, bb.width, bb.height, image->rows);
        WARNING(fuck);

        bb.x = 0;
        bb.width = image->columns;

        int diff_y, diff_height;

        if (bb.y > OVERLAP_RATIO * IMG_WIDTH / 2)
            diff_y = OVERLAP_RATIO * IMG_WIDTH / 2;
        else
            diff_y = bb.y;

        if (diff_y + bb.height < image->rows - OVERLAP_RATIO * IMG_WIDTH)
            diff_height = OVERLAP_RATIO * IMG_WIDTH;
        else // Can be negative
            diff_height = image->rows - bb.height - diff_y;

        bb.y -= diff_y;
        bb.height += diff_height;

        if (bb.y != 0 || bb.height != image->rows)
        {
            wsprintf(fuck, L"Cropping with y: %d, height: %d\n", bb.y, bb.height);
            WARNING(fuck);

            ResetImagePage(image, "0x0+0+0");
            start_y = bb.y;
            res_img = CropImage(image, &bb, &ex);
            WARNING(DLLTEXT("OK\n"));

            wsprintf(fuck, L"Resulting width: %d, height: %d\n", res_img->columns, res_img->rows);
            WARNING(fuck);
        }
        else
        {
            res_img = image;
        }
    }

    WARNING(DLLTEXT("Exiting GetCroppedImage.\n"));

    return res_img;
}

void ProcessImage(lUInt8* bitmap, DWORD& dwBufSize, DWORD& dwNewImgHeight,
                  DWORD width, DWORD height, BYTE btBPP,
                  BOOL bLandscape, BOOL bSplit, WOLWriter* pWolWriter,
                  BOOL bFlush)
{
    OEMDBG(DBG_VERBOSE, L"ProcessImage entry.");

    WARNING(DLLTEXT("Entered ProcessImage.\n"));

    ExceptionInfo ex;
    DWORD orig_width = width;
    lUInt8* bmp_orig = bitmap;
    lUInt8* bmp_new = NULL;

    if (btBPP == 1)
    {
        bmp_new = new lUInt8[width * height];

        for (int y = 0; y < height; ++y)
            for (int x = 0; x < width; ++x)
            {
                DWORD index = x + y * width;
                bmp_new[index] = (bitmap[index / 8] & (1 << (7 - index % 8)))? 255: 0;
            }
        bitmap = bmp_new;
    }

    if (bLandscape || bSplit)
    {
        DWORD dwByteBufSize = dwBufSize * 8 / btBPP;
        Image* image = ConstituteImage(width, height - dwByteBufSize / width, "I", CharPixel, bitmap + dwByteBufSize, &ex);

        DWORD real_img_start_y = 0;
        Image* img_crop;

        if (bSplit)
            img_crop = GetCroppedImage(image, real_img_start_y);
        else
            img_crop = image;

        if (img_crop || dwByteBufSize)
        {
            int prevHeight = dwByteBufSize / width;
            int nextHeight = img_crop? img_crop->rows : 0;
            int totalHeight = prevHeight + nextHeight;

            lUInt8* new_img = new lUInt8[totalHeight * width];
            memcpy(new_img, bitmap, dwByteBufSize);
            if (img_crop)
            {
                ExportImagePixels(img_crop, 0, 0, img_crop->columns, img_crop->rows, "I", CharPixel, new_img + dwByteBufSize, &ex);
                DestroyImage(img_crop);
            }

            if (img_crop != image)
                DestroyImage(image);

            image = ConstituteImage(width, totalHeight, "I", CharPixel, new_img, &ex);
            delete[] new_img;

            height = image->rows*IMG_HEIGHT/image->columns;
            width = IMG_HEIGHT;
            Image* img_res = ResizeImage(image, width, height, LanczosFilter, 1.0, &ex);
            DestroyImage(image);
            image = img_res;

            while (height >= IMG_WIDTH || bFlush && height > 0)
            {
                TCHAR fuck[255];
                wsprintf(fuck, L"Entering loop with height: %d, image rows: %d\n", height, image->rows);
                WARNING(fuck);

                RectangleInfo rect;
                rect.x = 0;
                rect.y = image->rows - height;
                rect.width = IMG_HEIGHT;
                rect.height = IMG_WIDTH;
                ResetImagePage(image, "0x0+0+0");
                Image* crop = CropImage(image, &rect, &ex);
                Image* image_rot = RotateImage(crop, 90, &ex);
                DestroyImage(crop);
                PosterizeImage(image_rot, 4, MagickTrue);
                if (image_rot->columns < IMG_WIDTH)
                {
                    rect.x = image_rot->columns - IMG_WIDTH;
                    rect.y = 0;
                    rect.width = IMG_WIDTH;
                    rect.height = IMG_HEIGHT;
                    Image* img_ext = ExtentImage(image_rot, &rect, &ex);
                    DestroyImage(image_rot);
                    image_rot = img_ext;
                }
                FlushImage(image_rot, pWolWriter);
                DestroyImage(image_rot);

                if (height <= IMG_WIDTH)
                    height = 0;
                else
                    height -= IMG_WIDTH * (1.0 - OVERLAP_RATIO);
            }

            TCHAR fuck[255];
            wsprintf(fuck, L"Adjusting buffer with image->rows: %d, height: %d, orig_width: %d\n", image->rows, height, orig_width);
            WARNING(fuck);

            DWORD row = (image->rows - height) * orig_width/IMG_HEIGHT + real_img_start_y;

            wsprintf(fuck, L"Calculated row: %d\n", row);
            WARNING(fuck);

            dwNewImgHeight = height * orig_width/IMG_HEIGHT;
            dwBufSize = dwNewImgHeight * orig_width * btBPP / 8;
            MoveMemory(bmp_orig, bmp_orig + orig_width * row * btBPP / 8, dwBufSize);

            WARNING(DLLTEXT("Done.\n"));
        }
        else
        {
            dwBufSize = 0;
            dwNewImgHeight = 0;
        }
        DestroyImage(image);
    }
    else
    {
        Image* image = ConstituteImage(width, height, "I", CharPixel, bitmap, &ex);
        height = height*IMG_WIDTH/width;
        width = IMG_WIDTH;
        Image* img_res = ResizeImage(image, width, height, LanczosFilter, 1.0, &ex);
        DestroyImage(image);
        image = img_res;
        PosterizeImage(image, 4, MagickTrue);
        FlushImage(image, pWolWriter);
        dwBufSize = 0;
        dwNewImgHeight = 0;
        DestroyImage(image);
    }

    if (bmp_new)
        delete[] bmp_new;

    WARNING(DLLTEXT("Exiting ProcessImage.\n"));
}
